
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnome.h>
#include <glib/gi18n.h>
#include <gconf/gconf-client.h>
#include <gdk/gdkx.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <libhal.h>
#include <libhal-storage.h>

static DBusConnection *dbus_connection;
static LibHalContext *hal_ctx;

/** Integrate a dbus mainloop. 
 *
 *  @param  ctx                 LibHal context
 *  @param  error     		pointer to a D-BUS error object
 *
 *  @return 			TRUE if we connected to the bus
 */
static dbus_bool_t 
hal_mainloop_integration (LibHalContext *ctx, DBusError *error)
{
}

/** Internal HAL initialization function
 *
 * @return			The LibHalContext of the HAL connection or
 *				NULL on error.
 */
static LibHalContext *
do_hal_init (void)
{
	LibHalContext *ctx;
	DBusError error;
	char **devices;
	int nr;
	
	ctx = libhal_ctx_new ();
	if (ctx == NULL) {
		g_warning ("Failed to get libhal context");
		goto error;
	}
	
	dbus_error_init (&error);
	dbus_connection = dbus_bus_get (DBUS_BUS_SYSTEM, &error);
	if (dbus_error_is_set (&error)) {
		g_warning ("Cannot connect to system bus: %s : %s", error.name, error.message);
		dbus_error_free (&error);
		goto error;
	}
	
        dbus_connection_setup_with_g_main (dbus_connection, NULL);	
	libhal_ctx_set_dbus_connection (ctx, dbus_connection);
	
	if (!libhal_ctx_init (ctx, &error)) {
		g_warning ("Failed to initialize libhal context: %s : %s", error.name, error.message);
		dbus_error_free (&error);
		goto error;
	}

	return ctx;

error:
	if (ctx != NULL)
		libhal_ctx_free (ctx);
	return NULL;
}

static gboolean
volume_mount (LibHalVolume *volume, const char *mount_point, const char *fstype, GList *fsoptions)
{
	gboolean ret = FALSE;
	unsigned int n;
	GList *i;
	unsigned int num_options;
	char **options = NULL;
	DBusMessage *msg = NULL;
	DBusMessage *reply = NULL;
	DBusError error;

	if (mount_point == NULL)
		mount_point = "";

	if (fstype == NULL)
		fstype = "";

	msg = dbus_message_new_method_call ("org.freedesktop.Hal", libhal_volume_get_udi (volume),
					    "org.freedesktop.Hal.Device.Volume",
					    "Mount");
	if (msg == NULL) {
		g_warning ("Could not create dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}

	num_options = g_list_length (fsoptions);
	options = g_new0 (char *, num_options + 1);

	///*
	g_message ("mounting %s with mount_point='%s', fstype='%s', num_options=%d", 
		   libhal_volume_get_udi (volume),
		   mount_point, 
		   fstype,
		   num_options);
	//*/

	for (i = fsoptions, n = 0; i != NULL; i = g_list_next (i), n++) {
		options[n] = (char *) i->data;
		g_message ("  option='%s'", options[n]);
	}
	options[n] = NULL;
	
	if (!dbus_message_append_args (msg, 
				       DBUS_TYPE_STRING, &mount_point, 
				       DBUS_TYPE_STRING, &fstype,
				       DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &options, num_options,
				       DBUS_TYPE_INVALID)) {
		g_warning ("Could not append args to dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}

	dbus_error_init (&error);
	reply = dbus_connection_send_with_reply_and_block (dbus_connection, msg, -1, &error);
	if (reply = NULL) {
		g_warning ("Mount failed for %s: %s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
		dbus_error_free (&error);
		goto out;
	}

	if (dbus_error_is_set (&error)) {
		g_warning ("Mount failed for %s\n%s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
		dbus_error_free (&error);
		goto out;
	}

	ret = TRUE;

out:
	if (msg != NULL)
		dbus_message_unref (msg);
	if (options != NULL)
		g_free (options);
	if (reply != NULL)
		dbus_message_unref (reply);

	return ret;
}

static gboolean
volume_unmount (LibHalVolume *volume)
{
	gboolean ret = FALSE;
	DBusMessage *msg = NULL;
	DBusMessage *reply = NULL;
	DBusError error;
	char **options = NULL;
	unsigned int num_options = 0;

	msg = dbus_message_new_method_call ("org.freedesktop.Hal", libhal_volume_get_udi (volume),
					    "org.freedesktop.Hal.Device.Volume",
					    "Unmount");
	if (msg == NULL) {
		g_warning ("Could not create dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}

	if (!dbus_message_append_args (msg, 
				       DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &options, num_options,
				       DBUS_TYPE_INVALID)) {
		g_warning ("Could not append args to dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}
	
	/*g_message ("unmounting %s", libhal_volume_get_udi (volume));*/
	
	dbus_error_init (&error);
	reply = dbus_connection_send_with_reply_and_block (dbus_connection, msg, -1, &error);
	if (reply = NULL) {
		g_warning ("Unmount failed for %s: %s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
		dbus_error_free (&error);
		goto out;
	}

	if (dbus_error_is_set (&error)) {
		g_warning ("Unmount failed for %s\n%s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
		dbus_error_free (&error);
		goto out;
	}

	ret = TRUE;

out:
	if (msg != NULL)
		dbus_message_unref (msg);
	if (reply != NULL)
		dbus_message_unref (reply);

	return ret;
}

static gboolean
volume_eject (LibHalVolume *volume)
{
	gboolean ret = FALSE;
	DBusMessage *msg = NULL;
	DBusMessage *reply = NULL;
	DBusError error;
	char **options = NULL;
	unsigned int num_options = 0;

	msg = dbus_message_new_method_call ("org.freedesktop.Hal", libhal_volume_get_udi (volume),
					    "org.freedesktop.Hal.Device.Volume",
					    "Eject");
	if (msg == NULL) {
		g_warning ("Could not create dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}

	if (!dbus_message_append_args (msg, 
				       DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &options, num_options,
				       DBUS_TYPE_INVALID)) {
		g_warning ("Could not append args to dbus message for %s", libhal_volume_get_udi (volume));
		goto out;
	}
	
	/*g_message ("ejecting %s", libhal_volume_get_udi (volume));*/
	
	dbus_error_init (&error);
	reply = dbus_connection_send_with_reply_and_block (dbus_connection, msg, -1, &error);
	if (reply = NULL) {
		g_warning ("Eject failed for %s: %s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
		dbus_error_free (&error);
		goto out;
	}

	if (dbus_error_is_set (&error)) {
		g_warning ("Eject failed for %s\n%s : %s\n", libhal_volume_get_udi (volume), error.name, error.message);
		dbus_error_free (&error);
		goto out;
	}

	ret = TRUE;

out:
	if (msg != NULL)
		dbus_message_unref (msg);
	if (reply != NULL)
		dbus_message_unref (reply);
	return ret;
}

int
main (int argc, char *argv[])
{
	LibHalVolume *volume = NULL;
	static gboolean opt_verbose = FALSE;
	static gboolean opt_unmount = FALSE;
	static gboolean opt_eject = FALSE;
	static gchar *opt_hal_udi = NULL;
	static gchar *opt_device_file = NULL;
        GError *error = NULL;
        GOptionContext *context = NULL;
        static GOptionEntry entries[] =
                {
                        { "verbose", 'v', 0, G_OPTION_ARG_NONE, &opt_verbose, "Verbose operation", NULL},
                        { "unmount", 'u', 0, G_OPTION_ARG_NONE, &opt_unmount, "Unmount rather than mount", NULL},
                        { "eject", 'e', 0, G_OPTION_ARG_NONE, &opt_eject, "Eject rather than mount", NULL},
                        { "hal-udi", 'h', 0, G_OPTION_ARG_STRING, &opt_hal_udi, "Mount by HAL UDI", NULL},
                        { "device", 'd', 0, G_OPTION_ARG_STRING, &opt_device_file, "Mount by device file", NULL},
                        { NULL, 0, 0, 0, NULL, NULL, NULL }
                };

	bindtextdomain (PACKAGE, GNOMELOCALEDIR);
	bind_textdomain_codeset (PACKAGE, "UTF-8");
	textdomain (PACKAGE);

        context = g_option_context_new ("- GNOME mount");
        g_option_context_add_main_entries (context, entries, PACKAGE);
        g_option_context_add_group (context, gtk_get_option_group (TRUE));
        g_option_context_parse (context, &argc, &argv, &error);

	gnome_program_init (PACKAGE, VERSION, LIBGNOMEUI_MODULE,
			    argc, argv, GNOME_PARAM_NONE);

	hal_ctx = do_hal_init ();
	if (hal_ctx == NULL)
		goto out;

	if (opt_hal_udi != NULL) {
		volume = libhal_volume_from_udi (hal_ctx, opt_hal_udi);
		if (volume == NULL) {
			g_message (_("The UDI '%s' is not a volume."), opt_hal_udi);
			goto out;
		}
	} else if (opt_device_file != NULL) {
		volume = libhal_volume_from_device_file (hal_ctx, opt_device_file);
		if (volume == NULL) {
			g_message (_("The device file '%s' is not a volume."), opt_device_file);
			goto out;
		}
	} else {
		g_message (_("Use --hal-udi or --device to specify volume"));
		goto out;
	}

	/* infer from commandline */
	if (strcmp (g_path_get_basename (argv[0]), "gnome-umount") == 0) {
		opt_unmount = TRUE;
	} else if (strcmp (g_path_get_basename (argv[0]), "gnome-eject") == 0) {
		opt_eject = TRUE;
	}

	if (opt_unmount && opt_eject) {
		g_message (_("Cannot unmount and eject simultaneously"));
		goto out;
	}

	if (opt_unmount) {
		g_message (_("Unmounting UDI '%s'"), libhal_volume_get_udi (volume));
		volume_unmount (volume);
	} else if (opt_eject) {
		g_message (_("Ejecting UDI '%s'"), libhal_volume_get_udi (volume));
		volume_eject (volume);
	} else {
		const char *fstype;

		fstype = libhal_volume_get_fstype (volume);
		if (fstype != NULL) {
			char uidbuf[64];
			char *mount_point;
			char *fstype_override;
			GList *fsoptions;

			fsoptions = NULL;

			snprintf (uidbuf, sizeof(uidbuf) - 1, "uid=%d", getuid());

			/* TODO: read defaults from gconf */
			if (strcmp (fstype, "vfat") == 0) {
				fsoptions = g_list_append (fsoptions, "shortname=winnt");
				fsoptions = g_list_append (fsoptions, uidbuf);
			} else if (strcmp (fstype, "iso9660") == 0) {
				fsoptions = g_list_append (fsoptions, uidbuf);
			} else if (strcmp (fstype, "udf") == 0) {
				fsoptions = g_list_append (fsoptions, uidbuf);
			}

			/* TODO: read per-volume settings from gconf and adjust fsoptions */

			/* TODO: read property volume.mount.valid_options and adjust fsoptions
			 *       (put into libhal-storage as a convenience function)
			 */

			/* TODO: read mount point from gconf */
			mount_point = NULL;

			/* TODO: read fstype override from gconf */
			fstype_override = NULL;

			g_message (_("Mounting UDI '%s'"), libhal_volume_get_udi (volume));
			volume_mount (volume, mount_point, fstype_override, fsoptions);

			g_list_free (fsoptions);
		}

	}
		
	/*gtk_main ();*/

out:	
 	if (volume != NULL)
		libhal_volume_free (volume);

	if (hal_ctx != NULL)
		libhal_ctx_free (hal_ctx);

	if (context != NULL)
		g_option_context_free (context);

	return 0;
}
